@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003-2016 GraphicsMagick Group
+% Copyright (C) 2003-2018 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 %
 % This program is covered by multiple licenses, which are described in
@@ -158,15 +158,18 @@ static Image *ReadMVGImage(const ImageInfo *image_info,ExceptionInfo *exception)
       /*
         Determine size of image canvas.
       */
+      (void) memset(&bounds,0,sizeof(bounds));
       while (ReadBlobString(image,primitive) != (char *) NULL)
       {
         for (p=primitive; (*p == ' ') || (*p == '\t'); p++);
         if (LocaleNCompare(BoundingBox,p,strlen(BoundingBox)) != 0)
           continue;
-        (void) sscanf(p,"viewbox %lf %lf %lf %lf",&bounds.x1,&bounds.y1,
-          &bounds.x2,&bounds.y2);
-        image->columns=(unsigned long) (bounds.x2-bounds.x1+0.5);
-        image->rows=(unsigned long) (bounds.y2-bounds.y1+0.5);
+        if (sscanf(p,"viewbox %lf %lf %lf %lf",&bounds.x1,&bounds.y1,
+                   &bounds.x2,&bounds.y2) == 4)
+          {
+            image->columns=(unsigned long) (bounds.x2-bounds.x1+0.5);
+            image->rows=(unsigned long) (bounds.y2-bounds.y1+0.5);
+          }
         break;
       }
     }
